#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#if defined(_WIN32)
#include <windows.h>
#endif

#include "License.h"

#include "Common.c"
#include "MD5.c"

#ifndef LIC_STATIC
#    define LIC_STATIC
#endif

LIC_STATIC void
licFree(void *p)
{
    if (p)
        free(p);
}

LIC_STATIC BOOL
licReadFile(char **license, char **err)
{
    char *home;
    int homelen;

    char *filename;
    struct stat statbuf;
    FILE *fp;
    char *buffer;
    int size;

    if (!licGetHomeDir(&home, &homelen, err))
        return FALSE;

    filename = (char*)malloc(homelen + 20);
    if (!filename) {
        free(home);
        return FALSE;
    }

    strcat(strcpy(filename, home), "ActiveState.lic");
    free(home);
    if (stat(filename, &statbuf) != 0) {
        *err = (char*)malloc(2048);
        if (*err)
            sprintf(*err, "License file '%s' does not exist", filename);
        free(filename);
        return FALSE;
    }

    size = statbuf.st_size + 10;
    *license = buffer = (char*)malloc(size);
    if (!buffer)
        return FALSE;

    fp = fopen(filename, "rt");
    if (!fp) {
        free(buffer);
        *err = (char*)malloc(2048);
        if (*err)
            sprintf(*err, "Cannot read license file '%s'", filename);
        free(filename);
        return FALSE;
    }

    while (size && fgets(buffer, size, fp)) {
        int len = strlen(buffer);
        if(buffer[len-2] == '\r') 
            buffer[len-2] = '\0';
        else buffer[len-1] = '\0';
        size -= len;
        buffer += len;
    }
    free(filename);
    fclose(fp);
    return TRUE;
}

static BOOL
Verify(const char *license, int len)
{
    char signature[33];
    if (!license || len < 34)
        return FALSE;
    MD5(license+33, signature);
    return memcmp(license, signature, 32) == 0;
}

LIC_STATIC char *
licFindFeature(const char *license, const char *feature)
{
    if (!license || !feature)
        return NULL;

    while (*license) {
        int len = strlen(license);
        if (Verify(license, len)) {
            const char *data = license+33;
            if (licCompareFeatures(feature, data) == 0)
                return strdup(data+strlen(feature));
        }
        license += len+1;
    }
    return NULL;
}

LIC_STATIC char *
licFindFirstNonExpired(const char *license, const char *key, const char *features[], char **name)
{
    if (!license || !features)
        return NULL;

    while (*features) {
        char *feature = licFindFeature(license, *features);
        if (feature) {
            int days;
            if (licDaysUntil(feature, key, &days) && days >= 0) {
                if (name)
                    *name = strdup(*features);
                return feature;
            }
            licFree(feature);
        }
        ++features;
    }
    return NULL;
}

LIC_STATIC char *
licFetchValue(const char *feature, const char *key)
{
    int len;
    char *buffer, *value, *end;

    if (!feature || !key)
        return NULL;

    len = strlen(key);
    buffer = (char*)malloc(len+3);
    sprintf(buffer, "|%s#", key);
    value = strstr(feature, buffer);
    free(buffer);
    if (!value)
        return NULL;

    value += len+2;
    end = strchr(value, '|');
    len = end ? end-value : strlen(value);
    buffer = (char*)malloc(len+1);
    memcpy(buffer, value, len);
    buffer[len] = '\0';
    return buffer;
}

static int
ymd_to_julian(int year, int month, int day)
{
    int adjust = 0;

    /* correct for negative year */
    if (year < 0)
	++year;

    if (month <= 2) {
	--year;
	month += 12;
    }

    /* deal with Gregorian calendar */
    if (year*10000. + month*100. + day >= 15821015.) {
	int a = year/100;
	adjust = 2 - a + a/4;
    }

    return (int)(365.25*year) +  (int)(30.6001 * (month+1)) + day +
           1720994L + adjust;
}

static int
YearMonthDayToDate(int year, int month, int day)
{
    static int daysInMonth[13] = { 0, 31, 28, 31, 30, 31, 30,
                                      31, 31, 30, 31, 30, 31 };
    BOOL leapYear;
    int daysInThisMonth;

    if (year < -4713 || month < 1 || month > 12)
	return -1;

    leapYear = (year >= 1582) ? ((year%4 == 0 && year%100 != 0) || year%400 == 0)
                              : (year%4 == 0);

    daysInThisMonth = daysInMonth[month];
    if (month == 2 && leapYear)
	++daysInThisMonth;

    if (day < 1 || day > daysInThisMonth)
	return -1;

    return ymd_to_julian(year, month, day);
}

LIC_STATIC BOOL
licDaysUntil(const char *feature, const char *key, int *days)
{
    time_t timebuf;
    struct tm *tmbuf;
    int today, expire;

    int i, day, month, year;
    char *date;

    if (!feature || !key)
        return FALSE;

    date = licFetchValue(feature, key);
    if (!date)
        return FALSE;

    /* Valid dates have the format "DD-MM-YYYY" */
    if (strlen(date) != 10) {
 Fail:
        free(date);
        return FALSE;
    }

    for (i=0; i<10; ++i) {
        if ((i==2 || i==5) ? date[i] != '-' : !isdigit(date[i]))
            goto Fail;
    }

    day   = atol(date);
    month = atol(date+3);
    year  = atol(date+6);
    free(date);

    if (day == 99 && month == 99 && year == 9999) {
        *days = DAYS_UNTIL_THE_END_OF_TIME;
        return TRUE;
    }

    if (day < 1 || day > 31 || month < 1 || month > 12 || year < 2001 || year > 2036)
        return FALSE;

    /* compute number of days until that day (may be negative) */
    expire = YearMonthDayToDate(year, month, day);
    time(&timebuf);
    tmbuf = localtime(&timebuf);
    today = YearMonthDayToDate(1900 + tmbuf->tm_year,
                               1 + tmbuf->tm_mon,
                               tmbuf->tm_mday);
    *days = expire - today;

    return TRUE;
}

LIC_STATIC void
licEnumFeatures(const char *license, BOOL (*callback)(void *any, char *name, char *data), void *any)
{
    if (!license)
        return;

    while (*license) {
        int len = strlen(license);
        if (Verify(license, len)) {
            BOOL abort;
            const char *payload = license+33;
            char *name = (char*)malloc(len-32);
            char *data = (char*)malloc(len-32);
            char *end = strchr(payload, '|');
            if (end) {
                memcpy(name, payload, end-payload);
                name[end-payload] = '\0';
                strcpy(data, end);
            }
            else {
                strcpy(name, payload);
                *data = '\0';
            }
            abort = (*callback)(any, name, data);
            free(name);
            free(data);
            if (abort)
                break;
        }
        license += len+1;
    }
}

LIC_STATIC void
licEnumKeys(const char *feature, BOOL (*callback)(void *any, char *key, char *value), void *any)
{
    int len;
    char *key, *value;

    if (!feature)
        return;

    len = strlen(feature);
    key = (char*)malloc(len+1);
    value = (char*)malloc(len+1);

    while (*feature) {
        const char *bar = strchr(feature+1, '|');
        const char *hash = strchr(feature+1, '#');

        if (!bar)
            bar = feature + strlen(feature);
        if (!hash)
            hash = feature + strlen(feature);

        len = hash-feature-1;
        memcpy(key, feature+1, len);
        key[len] = '\0';

        if (hash < bar) {
            len = bar-hash-1;
            memcpy(value, hash+1, len);
            value[len] = '\0';
        }
        else
            *value = '\0';

        if ((*callback)(any, key, value) || *bar == '\0')
            break;

        feature = bar;
    }
    free(key);
    free(value);
}
